/*
 * ops.c - Higher-level toolpath operations
 *
 * Written 2010 by Werner Almesberger
 * Copyright 2010 Werner Almesberger
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 */


#include <stddef.h>
#include <math.h>

#include "path.h"
#include "shape.h"
#include "ops.h"


static void tool_comp_1(struct path *path, int inside, int dog_bone)
{
	int left;
	struct path *new;

	left = path_tool_is_left(path);
	if (inside)
		new = path_offset(path, !left, path->notch);
	else
		new = path_offset(path, left, path->notch || dog_bone);
	path_replace(path, new);
}


void tool_comp_paths(struct path *paths, int dog_bone, int all_inside)
{
	struct path *leftmost, *path;

	/*
	 * We don't have an algorithm (yet) that can detect which paths are
	 * inside other paths. Therefore, we fake it by looking for the path
	 * that contains lowest x coordinate. This ought to be the outer
	 * boundary of the piece.
	 *
	 * Note that this heuristic falls apart when a job consists of
	 * multiple pieces. In this case, the #%outside hint can be used to
	 * explicitly tell cameo to treat the path as an outside edge.
	 */

	leftmost = path_find_leftmost(paths);
	for (path = paths; path; path = path->next)
		if (path != leftmost && (all_inside || !path->outside))
			tool_comp_1(path, 1, dog_bone);
	if (!all_inside)
		for (path = paths; path; path = path->next)
			if (path != leftmost && path->outside)
				tool_comp_1(path, 0, dog_bone);
	tool_comp_1(leftmost, all_inside, dog_bone);
}


struct path *try_drill(struct path *path, double d_min, double d_max)
{
	struct path *new;

	if (path->r_tool*2 < d_min || path->r_tool*2 > d_max)
		return NULL;
	if (!path->first || path->first != path->last)
		return NULL;
	new = path_new((d_min+d_max)/2); /* @@@ fishy */
	path_add(new, path->first->x, path->first->y, path->first->z);
	return new;
}


struct path *try_mill(struct path *path, double diam, double step)
{
	if (path->r_tool*2 < diam)
		return NULL;
	if (!path->first)
		return NULL;
	if (path->first == path->last)
		return circle(path->first->x, path->first->y, path->first->z,
		    path->r_tool, diam/2, step);
	if (path->first->next == path->last)
		return slot(path->first->x, path->first->y,
		    path->first->next->x, path->first->next->y,
		    path->first->z, path->r_tool, diam/2, step);
	return NULL;
}


/*
 * This isn't a perfect solution for the traveling salesman problem, but it's
 * easy to implement and usually produces results that don't look overly
 * offensive.
 */

struct path *optimize_paths(struct path *paths)
{
	struct path **walk, **best = NULL;
	struct path *res = NULL, **anchor = &res;
	struct path *curr;
	struct point *p;
	double best_d = 0, d;

	for (walk = &paths; *walk; walk = &(*walk)->next) {
		p = (*walk)->first;
		if (!p)
			continue;
		d = hypot(p->x, p->y);
		if (!best || d  < best_d) {
			best = walk;
			best_d = d;
		}
	}
	while (best) {
		curr = *best;
		*anchor = *best;
		anchor = &curr->next;
		*best = curr->next;
		best = NULL;
		for (walk = &paths; *walk; walk = &(*walk)->next) {
			p = (*walk)->first;
			if (!p)
				continue;
			d = hypot(p->x-curr->last->x, p->y-curr->last->y);
			if (!best || d < best_d) {
				best = walk;
				best_d = d;
			}
		}
	}
	return res;
}
